@@ -49,8 +49,29 @@ if "marshal" not in sys.builtin_module_names:
 import marshal
 
 
+#  Get the "time" module for debugging/profiling purposes.
+#  If it's not builtin, just use a dummy implementation.
+if "time" not in sys.builtin_module_names:
+    class time:
+        @staticmethod
+        def time():
+            return 0
+        @staticmethod
+        def clock():
+            return 0
+        @staticmethod
+        def besttime():
+            return 0
+else:
+    import time
+    if sys.platform == "win32":
+        time.besttime = time.clock
+    else:
+        time.besttime = time.time
+
+
 #  Get a minimal simulation of the "os" module.
-#  It must be builtin or this whole exercise is pointless.
+#  It must use only builtins or this whole exercise is pointless.
 def _signedimp_make_os_module():
     if "posix" in sys.builtin_module_names:
         from posix import stat
@@ -98,6 +119,7 @@ def _signedimp_make_os_module():
             @staticmethod
             def dirname(p):
                 """Local re-implementation of os.path.dirname."""
+                p = p.replace("/",SEP)
                 return SEP.join(p.split(SEP)[:-1])
     return os
 os = _signedimp_make_os_module()
@@ -105,6 +127,8 @@ os = _signedimp_make_os_module()
 
 HASHFILE_NAME = "signedimp-manifest.txt"
 
+signedimp_debug = False
+
 
 class IntegrityCheckError(Exception):
     """Error raised when integrity of a resource cannot be verified."""
@@ -193,25 +217,24 @@ class _signedimp_util:
             elif c == "/":
                 n += 63
             else:
-                raise ValueError
+                raise ValueError("invalid base64-encoded data")
         bytes = []
         while n > 0:
             bytes.append(chr(n & 0x000000FF))
             bytes.append(chr((n & 0x0000FF00) >> 8))
             bytes.append(chr((n & 0x00FF0000) >> 16))
             bytes.append(chr((n & 0xFF000000) >> 24))
-            n = n >> 32
+            n = (n & 0x00000000) >> 32
         while len(bytes) < 3:
             bytes.append("\x00")
-        bytes = "".join(reversed(bytes))
-        return bytes[-3:]
+        return bytes[2] + bytes[1] + bytes[0]
 
     @staticmethod
     def b64decode(data):
         """Pure-python base-64 decoder.
 
-        This is just awful and should only be used when the base64 module
-        is not available.
+        This is just awful.  But we decode so little base64 data that it's not
+        really worth trying to load up the native decoder.
         """
         output = []
         if len(data) % 4 != 0:
@@ -220,14 +243,10 @@ class _signedimp_util:
             quad = data[i:i+4]
             if quad.endswith("=="):
                 quad = quad[:2]+"AA"
-                trim = 2
+                output.append(_signedimp_util._b64unquad(quad)[:1])
             elif quad.endswith("="):
                 quad = quad[:3]+"A"
-                trim = 1
-            else:
-                trim = 0
-            if trim:
-                output.append(_signedimp_util._b64unquad(quad)[:-1*trim])
+                output.append(_signedimp_util._b64unquad(quad)[:2])
             else:
                 output.append(_signedimp_util._b64unquad(quad))
         return "".join(output)
@@ -240,12 +259,6 @@ class _signedimp_util:
         available.  It will attempt to import and use better implementations
         of the utility functions defined on this class.
         """
-        #  Try to use native b64decode
-        try:
-            import base64
-            _signedimp_util.b64decode = staticmethod(base64.b64decode)
-        except (ImportError,IntegrityCheckMissing):
-            pass
         #  Try to use our fast-path crypto library
         try:
             from signedimp.crypto import md5
@@ -280,44 +293,123 @@ class _signedimp_util:
                         pass
         #  If all else fails, we've left them as pure-python implementations
 
+    _timers = []
+    @staticmethod
+    def start_timer(msg,*args):
+        if __debug__ and signedimp_debug:
+            msg = msg % args
+            msg += " [%.2f secs]" % (time.clock(),)
+            _signedimp_util.debug(msg)
+            _signedimp_util._timers.append([time.besttime()])
+    @staticmethod
+    def checkpoint_timer(msg,*args):
+        if __debug__ and signedimp_debug:
+            now = time.besttime()
+            msg = msg % args
+            _signedimp_util._timers[-1].append(now)
+            msg += " [%.2f of %.2f secs]" % (now - _signedimp_util._timers[-1][-2],time.clock(),)
+            _signedimp_util.debug(msg)
+    @staticmethod
+    def stop_timer(msg,*args):
+        if __debug__ and signedimp_debug:
+            now = time.besttime()
+            tl = _signedimp_util._timers.pop()
+            msg = msg % args
+            msg += " [%.2f of %.2f secs]" % (now - tl[0],time.clock(),)
+            _signedimp_util.debug(msg)
+
+    @staticmethod
+    def profile_call(func):
+        if not __debug__ or not signedimp_debug:
+            return func
+        repr = _signedimp_util._reprobj
+        def wrapper(self,*args,**kwds):
+            argstr = ",".join(repr(a) for a in args)
+            argstr += "," + ",".join(k+"="+repr(v) for k,v in kwds.items())
+            _signedimp_util.start_timer("CALL> %s(%s)",func.func_name,argstr)
+            try:
+                return func(self,*args,**kwds)
+            finally:
+                _signedimp_util.stop_timer("CALL< %s(%s)",func.func_name,argstr)
+        return wrapper
+
+    @staticmethod
+    def _reprobj(obj):
+        if not isinstance(obj,basestring):
+            obj = repr(obj)
+        if len(obj) > 200:
+            obj = obj[:20] + "..."
+        return repr(obj)
+
+    @staticmethod
+    def debug(msg,*args):
+        """Print a debugging message to stderr, if enabled."""
+        if __debug__ and signedimp_debug:
+            msg = "  "*len(_signedimp_util._timers) + msg
+            if args:
+                msg = msg % args
+            sys.stderr.write(msg + "\n")
+            sys.stderr.flush()
+        
+
 
 
 class SignedHashDatabase(object):
     """An in-memory database of verified hash data.
 
-    This class is used to verify generic data blobs against a signed database
+    This class is used to verify file contents against a signed database
     of their expected hash values.  It encapsulates the file-parsing and
     verification logic shared by SignedImportManager and SignedLoader.
     """
 
-    def __init__(self,valid_keys=[],hashdata=None):
+    def __init__(self,valid_keys=[],hashdata=None,root_path=None):
         self.valid_keys = valid_keys
+        self.root_path = None
+        if root_path is not None:
+            self.root_path = self._normpath(root_path)
         self.hashes = {}
         if hashdata is not None:
             self.parse_hash_data(hashdata)
 
-    def verify(self,typ,name,data):
-        """Verify data of the given type against our hash database."""
+    def _normpath(self,path):
+        if sys.platform == "win32" and os.path.isabs(path):
+            if not path.startswith(os.sep):
+                path = path[2:]
+        path = path.replace(os.sep,"/")
+        if self.root_path is not None:
+            if path.startswith(self.root_path):
+                path = path[len(self.root_path):]
+                while path.startswith("/"):
+                    path = path[1:]
+        return path
+
+    @_signedimp_util.profile_call
+    def verify(self,path,data):
+        """Verify data for the given path against our hash database."""
+        path = self._normpath(path)
         try:
-            hashes = self.hashes[(typ,name)]
+            hashes = self.hashes[path]
         except KeyError:
-            raise IntegrityCheckMissing("no valid hash for "+name)
+            raise IntegrityCheckMissing("no valid hash for "+path)
         for (htyp,hval) in hashes:
             if not self._check_hash(htyp,hval,data):
-                raise IntegrityCheckFailed("invalid hash for "+name)
+                raise IntegrityCheckFailed("invalid hash for "+path)
 
-    def _check_hash(self,type,hash,data):
+    def _check_hash(self,typ,hash,data):
         """Check whether the hash of the given data matches the one given."""
-        if type == "sha1":
-            return _signedimp_util.sha1(data).hexdigest() == hash
-        if type == "md5":
-            return _signedimp_util.md5(data).hexdigest() == hash
-        raise ValueError("unknown hash type: %s" % (type,))
+        if typ == "sha1":
+            h_sha1 = _signedimp_util.sha1(data)
+            return (h_sha1.hexdigest() == hash)
+        elif typ == "md5":
+            h_md5 = _signedimp_util.md5(data)
+            return (h_md5.hexdigest() == hash)
+        raise ValueError("unknown hash type: %s" % (typ,))
 
     def _strip(self,s):
         """Compatability wrapper for compiling string.strip() under pypy."""
         return s.strip("\n").strip("\r").strip(" ")
 
+    @_signedimp_util.profile_call
     def parse_hash_data(self,hashdata):
         """Load hash data from the given string.
 
@@ -330,9 +422,9 @@ class SignedHashDatabase(object):
           key2fingerprint base64-signature2
 
           md5
-          m 76f3f13442c26fd4f1c709c7b03c6b76 os
-          m f56dbc5ee6774e857a7ef07accdbd19b hashlib
-          d 43b74fc5d2acb6b4e417f4feff06dd81 some/data/file.txt
+          76f3f13442c26fd4f1c709c7b03c6b76 os.pyc
+          f56dbc5ee6774e857a7ef07accdbd19b hashlib.pyc
+          43b74fc5d2acb6b4e417f4feff06dd81 some/data/file.txt
           ----
 
         """
@@ -354,6 +446,7 @@ class SignedHashDatabase(object):
                 for k in self.valid_keys:
                     if k.fingerprint() == fingerprint:
                         signatures.append((k,signature))
+                        break
         #  If there weren't any usable signatures, we can't use this data.
         if not signatures:
             return
@@ -372,15 +465,14 @@ class SignedHashDatabase(object):
         for ln in lines:
             try:
                 comps = self._strip(ln).split(" ")
-                typ = comps[0]
-                hval = comps[1]
-                name = " ".join(comps[2:])
+                hval = comps[0]
+                path = " ".join(comps[1:])
             except ValueError:
                 continue
             try:
-                hashes = self.hashes[(typ,name)]
+                hashes = self.hashes[path]
             except KeyError:
-                self.hashes[(typ,name)] = hashes = []
+                self.hashes[path] = hashes = []
             hashes.append((htyp,hval))
 
 
@@ -407,7 +499,12 @@ class SignedImportManager(object):
         self.reinstall()
 
     def install(self):
-        """Install this manager into the process import machinery."""
+        """Install this manager into the process-wide import machinery.
+
+        This method places the import manager as first entry on sys.meta_path,
+        and replaces some methods from the "imp" module with compatability
+        wrappers that call into the manager.
+        """
         if self not in sys.meta_path:
             sys.meta_path.insert(0,self)
             self._orig_load_dynamic = imp.load_dynamic
@@ -423,6 +520,7 @@ class SignedImportManager(object):
         #  Try to speed things up by loading faster crypto primitives.
         _signedimp_util.recreate()
 
+    @_signedimp_util.profile_call
     def find_module(self,fullname,path=None):
         """Get the loader for the given module.
 
@@ -430,6 +528,7 @@ class SignedImportManager(object):
         given module, and wraps it in a SignedLoader instance so that all
         data is verified immediately prior to being loaded.
         """
+        cached = True
         if path is not None and fullname.startswith("signedimp."):
             # Careful now, the signedimp module may have been created by
             # hand and its __path__ may not reflect the real sys.path.
@@ -439,10 +538,11 @@ class SignedImportManager(object):
                     path = []
                     for p in sys.path:
                         path.append(os.path.join(p,"signedimp"))
-        loader = self._find_loader(fullname,path)
+                cached = False
+        loader = self._find_loader(fullname,path,cached=cached)
         return SignedLoader(self,loader)
 
-    def _find_loader(self,fullname,path):
+    def _find_loader(self,fullname,path,cached=True):
         """Find the loader that would normally be used for the given module.
 
         This basically emulates the standard lookup machinery defined by PEP
@@ -460,8 +560,9 @@ class SignedImportManager(object):
             loader = BuiltinImporter.find_module(fullname)
             if loader is not None:
                 return loader
-            path = sys.path
-        #  Otherwise, we have to go looking along the given path
+        #  Try the items on sys.meta_path.  Only those appearing after this
+        #  object are tried, as anything before it must already have been
+        #  invoked and failed to load the module.
         found_me = False
         for mphook in sys.meta_path:
             if found_me:
@@ -470,36 +571,36 @@ class SignedImportManager(object):
                     return loader
             elif mphook is self:
                 found_me = True
+        # Try the items on the given path (or sys.path if not specified)
+        if path is None:
+            path = sys.path
         for pathitem in path:
-            importer = self._get_importer(pathitem)
+            importer = self._get_importer(pathitem,cached=cached)
             loader = importer.find_module(fullname)
             if loader is not None:
                 return loader
         raise ImportError(fullname)
 
-    def _get_importer(self,path):
+    def _get_importer(self,path,cached=True):
         """Get the importer for the given sys.path item.
 
         This emulates the standard handling of sys.path_hooks, with the added
         bonus of returning a DefaultImporter instance if no hook is found.
         """
-        try:
-            importer = sys.path_importer_cache[path]
-        except KeyError:
+        importer = None
+        if cached:
+            importer = sys.path_importer_cache.get(path,None)
+        if importer is None:
             for importer_class in sys.path_hooks:
                 try:
                     importer = importer_class(path)
                 except ImportError:
                     pass
                 else:
-                    sys.path_importer_cache[path] = importer
                     break
             else:
-                sys.path_importer_cache[path] = None
-                importer = _get_default_importer(path)
-        else:
-            if importer is None:
                 importer = _get_default_importer(path)
+            sys.path_importer_cache[path] = importer
         return importer
         
     def get_canonical_modname(self,fullname):
@@ -527,6 +628,7 @@ class SignedImportManager(object):
         resolved.append(self.module_aliases[unresolved])
         return ".".join(reversed(resolved))
         
+    @_signedimp_util.profile_call
     def load_hashdb(self,loader,path):
         """Load the hashdb at the given path in the given loader.
         
@@ -542,7 +644,8 @@ class SignedImportManager(object):
                 hashdb = self._hashdb_cache[path]
             except KeyError:
                 hashdata = loader.get_data(path)
-                hashdb = SignedHashDatabase(self.valid_keys)
+                root_path = os.path.dirname(path)
+                hashdb = SignedHashDatabase(self.valid_keys,root_path=root_path)
                 hashdb.parse_hash_data(hashdata)
                 self._hashdb_cache[path] = hashdb
         return hashdb
@@ -565,7 +668,8 @@ class SignedImportManager(object):
                     hashdata = f.read()
                 finally:
                     f.close()
-                hashdb = SignedHashDatabase(self.valid_keys)
+                root_path = os.path.dirname(hashfile)
+                hashdb = SignedHashDatabase(self.valid_keys,root_path=root_path)
                 hashdb.parse_hash_data(hashdata)
                 self._hashdb_cache[hashfile] = hashdb
                 return path,hashdb
@@ -573,7 +677,21 @@ class SignedImportManager(object):
             if path == new_path:
                 break
             path = new_path
+        return None,None
 
+    def _verify_file(self,pathname):
+        """Verify a file found on the local filesystem"""
+        (basepath,hashdb) = self._find_hashdb(pathname)
+        if hashdb is None:
+            raise IntegrityCheckMissing("no hashdb for " + pathname)
+        f = open(pathname,"rb")
+        try:
+            data = f.read()
+        finally:
+            f.close()
+        hashdb.verify(pathname,data)
+
+    @_signedimp_util.profile_call
     def _imp_load_dynamic(self,name,pathname,file=None):
         """Replacement for imp.load_dynamic.
         
@@ -581,23 +699,13 @@ class SignedImportManager(object):
         function, checking signatures before calling back to the original
         implementation.
         """
-        (basepath,hashdb) = self._find_hashdb(pathname)
-        if hashdb is None:
-            raise IntegrityCheckMissing(pathname)
-        f = open(pathname,"rb")
-        try:
-            data = f.read()
-        finally:
-            f.close()
-        try:
-            hashdb.verify("m",name,data)
-        except IntegrityCheckMissing:
-            hashdb.verify("d",pathname[len(basepath)+1:],data)
+        self._verify_file(pathname)
         if file is not None:
             return self._orig_load_dynamic(name,pathname,file)
         else:
             return self._orig_load_dynamic(name,pathname)
 
+    @_signedimp_util.profile_call
     def _imp_load_compiled(self,name,pathname,file=None):
         """Replacement for imp.load_compiled.
         
@@ -605,23 +713,13 @@ class SignedImportManager(object):
         function, checking signatures before calling back to the original
         implementation.
         """
-        (basepath,hashdb) = self._find_hashdb(pathname)
-        if hashdb is None:
-            raise IntegrityCheckMissing(pathname)
-        f = open(pathname,"rb")
-        try:
-            data = f.read()
-        finally:
-            f.close()
-        try:
-            hashdb.verify("m",name,data)
-        except IntegrityCheckMissing:
-            hashdb.verify("d",pathname[len(basepath)+1:],data)
+        self._verify_file(pathname)
         if file is not None:
             return self._orig_load_compiled(name,pathname,file)
         else:
             return self._orig_load_compiled(name,pathname)
 
+    @_signedimp_util.profile_call
     def _imp_load_source(self,name,pathname,file=None):
         """Replacement for imp.load_source.
         
@@ -629,18 +727,7 @@ class SignedImportManager(object):
         function, checking signatures before calling back to the original
         implementation.
         """
-        (basepath,hashdb) = self._find_hashdb(pathname)
-        if hashdb is None:
-            raise IntegrityCheckMissing(pathname)
-        f = open(pathname,"rb")
-        try:
-            data = f.read()
-        finally:
-            f.close()
-        try:
-            hashdb.verify("m",name,data)
-        except IntegrityCheckMissing:
-            hashdb.verify("d",pathname[len(basepath)+1:],data)
+        self._verify_file(pathname)
         if file is not None:
             return self._orig_load_source(name,pathname,file)
         else:
@@ -677,31 +764,31 @@ class SignedLoader:
             raise AttributeError("unsafe attribute type: %s" % (attr,))
         return value
 
+    @_signedimp_util.profile_call
     def load_module(self,fullname,verify=True):
         """Load the specified module, checking its integrity first.
 
         This method is really key to the whole apparatus - it requests the
         raw module data from the loader object, verifies it against the
-        hash database, and only if it's valid does it request that the loader
-        actually import the module.
+        hash database, and only if it's valid does it proceed to load the
+        module.  Where possible, the module is loaded without re-reading
+        the data from disk.
 
         If you really want to load non-verified code, you can pass the kwd arg
         "verify" as False.  But seriously, why would you want to do that?
         """
-        try:
-            return sys.modules[fullname]
-        except KeyError:
-            pass
-        if verify and self.loader is not BuiltinImporter:
-            self.verify_module(fullname)
-        mod = self.loader.load_module(fullname)
+        if not verify or self.loader is BuiltinImporter:
+            mod = self.loader.load_module(fullname)
+        else:
+            mod = self._load_verified_module(fullname)
         mod.__loader__ = self
         return mod
 
+    @_signedimp_util.profile_call
     def get_data(self,path):
         """Get the data from the given path, checking its integrity first."""
         data = self.loader.get_data(path)
-        self._verify("d",path,data)
+        self._verify(path,data)
         return data
 
     def is_package(self,fullname):
@@ -734,9 +821,16 @@ class SignedLoader:
         """
         return self.loader.get_source(fullname)
 
+    @_signedimp_util.profile_call
     def get_filename(self,fullname):
         """Get the filename associated with the given module."""
-        return self.loader.get_filename(fullname)
+        try:
+            return self.loader.get_filename(fullname)
+        except AttributeError:
+            try:
+                return self.loader._get_filename(fullname)
+            except AttributeError:
+                raise AttributeError("get_filename")
 
     def get_datafilepath(self,path):
         """Get the full path for the given data file.
@@ -757,84 +851,120 @@ class SignedLoader:
         else:
             return gdfp(path)
 
-    def verify_module(self,fullname):
-        """Verify the given module's data against our signature database."""
-        data = self._get_module_data(fullname)
-        if self.is_package(fullname):
-            valname = fullname + ".__init__"
-        else:
-            valname = fullname
-        try:
-            self._verify("m",valname,data)
-        except IntegrityCheckMissing:
-            cname = self.manager.get_canonical_modname(valname)
-            if cname == valname:
-                raise
-            data = self._get_module_data(cname)
-            self._verify("m",cname,data)
-
-    def _verify(self,typ,name,data):
-        """Verify date for the given type specifier and name.
+    @_signedimp_util.profile_call
+    def _verify(self,path,data,canonicalise=True):
+        """Verify data for the given path.
 
         This performs verification against the local database and the main
         manager database.  If a valid hash for the item is not found in
         either location, IntegrityCheckMissing error is raised.
         """
         try:
-            self.hashdb.verify(typ,name,data)
-        except IntegrityCheckMissing:
-            self.manager.hashdb.verify(typ,name,data)
-        else:
             try:
-                self.manager.hashdb.verify(typ,name,data)
+                self.hashdb.verify(path,data)
             except IntegrityCheckMissing:
-                pass
+                self.manager.hashdb.verify(path,data)
+            else:
+                try:
+                    self.manager.hashdb.verify(path,data)
+                except IntegrityCheckMissing:
+                    pass
+        except IntegrityCheckMissing:
+            if not canonicalise:
+                raise
+            for (suffix,_,typ) in imp.get_suffixes():
+                if path.endswith(suffix):
+                    modname = path[:-1*len(suffix)]
+                    modname = modname.replace("/",".").replace("\\",".")
+                    cmodname = self.manager.get_canonical_modname(modname)
+                    if cmodname != modname:
+                        cpath = cmodname.replace(".","/") + suffix
+                        self._verify(cpath,data,False)
+                        break
+            else:
+                raise
 
-    def _get_module_data(self,fullname):
-        """Get the raw data for the given module.
+    def _load_verified_module(self,fullname):
+        """Verify and load the named module.
 
-        This is the data that must be included in a valid signature.  It's
-        a concatentation of any of the following that are found, searched for
-        in order:
-
-            * the source code from the module source file
-            * the raw data from the compiled bytecode file
-            * the raw data from the object file
-
-        This assumes that the loader has files laid out in the same scheme
-        as the default import machinery, so we can convert a module name to
-        a path and read the file using loader.get_data().
+        This assumes that the wrapped loader has files laid out in the same
+        scheme as the default import machinery, so we can convert a module name
+        to a path and read the file using loader.get_data().
         """
+        impname = fullname
+        is_package = False
         if self.is_package(fullname):
-            fullname = fullname + ".__init__"
-        modname = fullname.rsplit(".",1)[-1]
-        #  Try to use the final value of __file__ to determine the files to read.
+            impname += ".__init__"
+            is_package = True
+        modname = impname.rsplit(".",1)[-1]
+        #  Try to use the final value of __file__ to determine files to read.
         #  This works better for packages that play tricks with __path__.
         #  Otherwise, make the filenames by replacing "." with "/" in fullname.
         try:
             fn = self.get_filename(fullname)
             dirnm = os.path.dirname(fn)
-        except (ImportError,AttributeError):
-            dirnm = None
-        data = []
-        for (suffix,_,_) in imp.get_suffixes():
+        except (ImportError,AttributeError), e:
+            dirnm = fn = None
+        data_found = False
+        #  Try each canidate file path in turn.
+        for (suffix,_,typ) in imp.get_suffixes():
             for sep in ("/","\\"):
                 if dirnm is None:
-                    path = fullname.replace(".",sep) + suffix
+                    path = impname.replace(".",sep) + suffix
                 else:
                     path = dirnm + sep + modname + suffix
                 try:
-                    new_data = self.loader.get_data(path)
+                    data = self.loader.get_data(path)
                 except AttributeError:
                     err = "loader has no get_data method"
                     raise IntegrityCheckMissing(err)
                 except IOError:
                     pass
                 else:
-                    data.append(new_data)
+                    #  Verify the data using a canonicalised path.
+                    self._verify(path,data)
+                    data_found = True
                     break
-        return "\x00".join(data)
+            else:
+                continue
+            #  If it's data that we can turn into a module by ourselves,
+            #  do so and avoid re-reading it from disk.  Note that this
+            #  defeats the use of cached bytecode files.
+            if typ in (imp.PY_SOURCE,imp.PY_COMPILED):
+                if typ == imp.PY_COMPILED:
+                    code = marshal.loads(data[8:])
+                elif typ == imp.PY_SOURCE:
+                    data = data.replace("\r\n","\n")
+                    code = compile(data,path,"exec")
+                mod = self._create_module(fullname,code,path,is_package)
+                break
+        else:
+            #  It must be something we can't load ourselves.  All checks have
+            #  passed so just get the loader to read it from disk.
+            if not data_found:
+                raise ImportError("no data found for: " + fullname)
+            mod = self.loader.load_module(fullname)
+        return mod
 
+    def _create_module(self,fullname,code,filename,is_package):
+        """Create a new module by executing the given code object."""
+        created = False
+        mod = sys.modules.get(fullname)
+        if not mod:
+            mod = imp.new_module(fullname)
+            sys.modules[fullname] = mod
+            created = True
+        try:
+            mod.__file__ = filename
+            mod.__loader__ = self
+            if is_package:
+                mod.__path__ = [os.path.dirname(filename)]
+            exec code in mod.__dict__
+        except Exception:
+            if created:
+                sys.modules.pop(fullname)
+            raise
+        return mod
 
 
 class BuiltinImporter(object):
@@ -876,6 +1006,14 @@ class BuiltinImporter(object):
             return True
         return False
 
+    @classmethod
+    def get_code(self,fullname):
+        return None
+
+    @classmethod
+    def get_source(self,fullname):
+        return None
+
 
 
 def _get_default_importer(path=None):
@@ -1069,21 +1207,23 @@ class DefaultImporter:
         """
         if self.base_path is None:
             raise OSError
-        f = open(os.path.join(self.base_path,path),"rb")
+        path = self._normpath(path)
+        f = open(path,"rb")
         try:
             return f.read()
         finally:
             f.close()
 
     def get_filename(self,fullname):
-        """Get the code object for the given module."""
+        """Get the filename assigned to the given module."""
         file,pathname,description = self._get_module_info(fullname)
         if file is not None:
             file.close()
         if description[2] == imp.PKG_DIRECTORY:
-            pkgpath = os.path.join(self.path,fullname.rsplit(".",1)[1])
-            importer = _get_default_importer(pkgpath)
-            return importer.get_filename(fullname + ".__init__")
+            for nm in ("__init__.pyc","__init__.py"):
+                if os.path.exists(os.path.join(pathname,nm)):
+                    return os.path.join(pathname,nm)
+            raise ImportError("no module named " + fullname)
         else:
             return pathname
 
@@ -1091,6 +1231,12 @@ class DefaultImporter:
         """Get the full path for the given data file."""
         if self.base_path is None:
             return None
+        return self._normpath(path)
+
+    def _normpath(self,path):
+        if not os.path.isabs(self.base_path):
+            if path.startswith(self.base_path+os.sep):
+                return path[len(self.base_path)+1:]
         return os.path.join(self.base_path,path)
 
 
