@@ -43,8 +43,25 @@ from signedimp.crypto.sha1 import sha1
 from signedimp.crypto.md5 import md5
 from signedimp.crypto.rsa import RSAKey
 
-if sys.platform == "win32":
-    from signedimp import winres
+from signedimp import refreeze
+
+
+def _get_source_lines(mod,indent,inline_crypto=True):
+    if isinstance(mod,basestring):
+        mod = __import__(mod,fromlist=["*"])
+    src = inspect.getsource(mod)
+    for ln in src.split("\n"):
+        if inline_crypto:
+            if ln.strip().startswith("from signedimp.cryptobase."):
+                lnstart = ln.find("from")
+                newindent = indent + ln[:lnstart]
+                newmod = ln.strip()[5:].split()[0]
+                for newln in _get_source_lines(newmod,newindent):
+                    yield newln
+            else:
+                yield indent + ln
+        else:
+            yield indent + ln
 
 
 def get_bootstrap_code(indent=""):
@@ -66,18 +83,6 @@ def get_bootstrap_code(indent=""):
        ''' % (signedimp.tools.get_bootstrap_code(),)
 
     """
-    def _get_source_lines(mod,indent):
-        mod = __import__(mod,fromlist=["*"])
-        src = inspect.getsource(mod)
-        for ln in src.split("\n"):
-            if ln.strip().startswith("from signedimp.cryptobase."):
-                lnstart = ln.find("from")
-                newindent = indent + ln[:lnstart]
-                newmod = ln.strip()[5:].split()[0]
-                for newln in _get_source_lines(newmod,newindent):
-                    yield newln
-            else:
-                yield indent + ln
     return """
 %(indent)ssignedimp = sys.modules.get("signedimp",None)
 %(indent)sif signedimp is None:
@@ -286,34 +291,21 @@ except (IndexError,AttributeError):
     signedimp.SignedImportManager([k]).install()
 
 """ % locals()
-    bscode = compile(bscode,"__main__.py","exec")
     #  Hack the bootstrap code into the start of each script to be run.
-    #  This unfortunately depends on some inner details of the py2exe format.
     for nm in os.listdir(appdir):
         if nm.endswith(".exe"):
             exepath = os.path.join(appdir,nm)
             try:
-                appcode = winres.load_resource(exepath,u"PYTHONSCRIPT",1,0)
+                app = refreeze.Py2Exe(exepath)
             except EnvironmentError:
-                continue
-            sz = struct.calcsize("iiii")
-            (magic,optmz,bfrd,codelen) = struct.unpack("iiii",appcode[:sz])
-            assert magic == 0x78563412
-            codebytes = appcode[sz:-1]
-            for i,c in enumerate(codebytes):
-                if c == "\x00":
-                    relarcname = codebytes[:i]
-                    codelist = marshal.loads(codebytes[i+1:-1])
-                    break
-            codelist.insert(0,bscode)
-            codebytes = marshal.dumps(codelist)
-            appcode = struct.pack("iiii",magic,optmz,bfrd,len(codebytes)) \
-                      + relarcname + "\x00" + codebytes + "\x00\x00"
-            winres.add_resource(exepath,appcode,u"PYTHONSCRIPT",1,0)
+                pass
+            else:
+                app.prepend_code(bscode)
     #  Sign anything that might be an importable zipfile.
     for nm in os.listdir(appdir):
         if nm.endswith(".exe") or nm.endswith(".zip"):
             try:
+
                 sign_zipfile(os.path.join(appdir,nm),key,hash=hash)
             except zipfile.BadZipfile:
                 pass
@@ -384,12 +376,8 @@ except (IndexError,AttributeError):
     signedimp.SignedImportManager([k]).install()
 
 """ % locals()
-    bsfile = os.path.join(appdir,"Contents","Resources","__boot__.py")
-    with open(bsfile,"r+") as f:
-        oldcode = f.read()
-        f.seek(0)
-        f.write(bscode)
-        f.write(oldcode)
+    app = refreeze.Py2App(appdir)
+    app.prepend_code(bscode)
     #  Sign the main library.zip
     libdir = os.path.join(appdir,"Contents","Resources","lib")
     libdir = os.path.join(libdir,"python%d.%d"%sys.version_info[:2])
@@ -452,7 +440,7 @@ def sign_cxfreeze_app(appdir,key=None,hash="sha1",check_modules=None):
     #  Build the bootstrap code to be inserted into each executable.  Since
     #  it replaces the cx_Freeze__init__ script it needs to exec that once
     #  the signed imports are in place.
-    bscode_tmplt =  """
+    bscode =  """
 import sys
 
 #  Check the boot-time modules if necessary.
@@ -481,47 +469,16 @@ try:
         signedimp.SignedImportManager([k]).install()
 except (IndexError,AttributeError):
     signedimp.SignedImportManager([k]).install()
-
-#  Bootstrap the original cx_Freeze__init__ module.
-#  If it was in the appended zipfile, we're given it as a marshalled string.
-#  If not, we need to search for it in the other zipfiles.
-if %(has_initcode)r:
-    import marshal
-    exec marshal.loads(%(initcode)r)
-else:
-    import zipimport
-    initmod = "cx_Freeze__init__"
-    try:
-        imp = zipimport.zipimporter(EXCLUSIVE_ZIP_FILE_NAME)
-        imp.find_module(initmod)
-        INITSCRIPT_ZIP_FILE_NAME = EXCLUSIVE_ZIP_FILE_NAME
-    except ImportError:
-        imp = zipimport.zipimporter(SHARED_ZIP_FILE_NAME)
-        imp.find_module(initmod)
-        INITSCRIPT_ZIP_FILE_NAME = SHARED_ZIP_FILE_NAME
-    code = imp.get_code(initmod)
-    exec code
-    
-"""
+""" % locals()
     #  Add the bootstrapping code to any executables found in the dir.
     for nm in os.listdir(appdir):
         fpath = os.path.join(appdir,nm)
         if not os.path.isfile(fpath) or not _is_executable(fpath):
             continue
+        app = refreeze.CXFreeze(fpath)
+        app.prepend_code(bscode)
         zf = zipfile.PyZipFile(fpath,"a")
         try:
-            #  If it contains the init module already, we'll need to
-            #  grab its code to bootstrap into it.
-            try:
-                initcode = repr(zf.read(initmod+".pyc")[8:])
-            except KeyError:
-                initcode = ""
-            #  Store our own code as the cxfreeze init module
-            has_initcode = bool(initcode)
-            bssrc = bscode_tmplt % locals()
-            bscode = imp.get_magic() + struct.pack("<i",time.time())
-            bscode += marshal.dumps(compile(bssrc,initmod+".py","exec"))
-            zf.writestr(initmod+".pyc",bscode)
             #  Make sure the signedimp module is bundled into the zipfile
             zf.writepy(os.path.dirname(signedimp.__file__))
             #  The python interpreter itself tries to import various encodings
@@ -566,28 +523,6 @@ def _is_executable(path):
         return ("executable" in p.stdout.read())
 
 
-def _is_in_package(root,path,os=os):
-    """Check whether the given path is inside a package directory structure.
-
-    This is used to decide whether to sign it as a module or a datafile.
-    """
-    while path != root:
-        for (suffix,_,_) in imp.get_suffixes():
-            if os.path.exists(os.path.join(path,"__init__"+suffix)):
-                break
-        else:
-            return False
-        path = os.path.dirname(path)
-    return True
-
-
-def _get_module_basename(filepath,os=os):
-    """Get the base name of the module at the given file path."""
-    for (suffix,_,_) in imp.get_suffixes():
-        if filepath.endswith(suffix):
-            return filepath[:-1*len(suffix)]
-    return None
-
 def _read_file(path):
     """Default read() function for use with hash_files()."""
     with open(path,"rb") as f:
@@ -608,7 +543,6 @@ def hash_files(path,files=None,hash="sha1",read=_read_file,os=os):
         hash = md5
     else:
         raise ValueError("unknown hash type: %s" % (hash,))
-    modhashes = {}
     datahashes = {}
     while path.endswith(os.sep):
         path = path[:-1]
@@ -623,36 +557,12 @@ def hash_files(path,files=None,hash="sha1",read=_read_file,os=os):
                     yield os.path.join(dirnm,filenm)
         files = files()
     for filepath in files:
-        #  If we're not in a package, sign everything as a datafile.
-        #  If we are, try to sign files as a module first.
-        if not _is_in_package(path,os.path.dirname(filepath),os=os):
-            #  Just sign it as a datafile
-            hashname = filepath[prefixlen:].replace(os.sep,"/")
-            datahashes[hashname] = hash(read(filepath)).hexdigest()
-        else:
-            basenm = _get_module_basename(filepath,os=os)
-            if basenm is None:
-                #  Just sign it as a datafile
-                hashname = filepath[prefixlen:].replace(os.sep,"/")
-                datahashes[hashname] = hash(read(filepath)).hexdigest()
-            else:
-                #  We sign the concatentation of all files in the order
-                #  they are found by the import machinery.
-                modname = basenm[prefixlen:].replace(os.sep,".")
-                if modname not in modhashes:
-                    moddata = []
-                    for (suffix,_,_) in imp.get_suffixes():
-                        if os.path.exists(basenm+suffix):
-                            modpath = basenm+suffix
-                            moddata.append(read(modpath))
-                    moddata = "\x00".join(moddata)
-                    modhashes[modname] = hash(moddata).hexdigest()
+        hashname = filepath[prefixlen:].replace(os.sep,"/")
+        datahashes[hashname] = hash(read(filepath)).hexdigest()
     #  Concatenate the various hashes.  Put then in sorted order so
     #  it's easy to locate a particular hash by hand.
-    for nm,hash in sorted(modhashes.items()):
-        output.append("m %s %s" % (hash,nm))
     for nm,hash in sorted(datahashes.iteritems()):
-        output.append("d %s %s" % (hash,nm))
+        output.append("%s %s" % (hash,nm))
     return "\n".join(output)
 
 
